require("debug_util")

--[[
	>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	Version: 0.0.1 - Build: 180501 - Factorio 0.16.36
	
	
    Description:
	-------------------------
	You love trains in Factorio? , You don't like loops?, You don't like the look/feel of dual-header trains? You have a beauty & symmetry OCD? 
	Or you sometimes just need to turn around your personal train?
	
	Then maybe Train Engineer is a mod for YOU!
	
	With this mod you get a skilled and fully certified Train Engineer named Bob (see https://en.wikipedia.org/wiki/Railroad_engineer). After years of ative duty
	Train Engineer Bob has become so skilled his actions seem like magic. He can shunt/switch trains at terminal stations without another track or locomotive. 
	And he works currently for free.
	
	example for progress of events:
	-------------------------------
	
	    terminal station
	1) =LL-CCCCCCC============== >> train arrives at station (inbound movement)
	
	2) =LL-CCCCCCC============== >> train waits until leave conditions are met
	
	3) Train Engineer Bob does his magic
	
	    terminal station
	4) =CCCCCCC-LL============== >> train leaves station (outbound movement)
	
	During event 3) to 4):
	- mod reads complete train and copies certain parameters (train composition, carriage types, cargo inventory, fuel of locomotives, loco colors, schedule, train_mode)
	- old train gets removed
	- new train gets placed in reversed direction + parameters and inventory is inserted (a completely new train with a new Train id is created!)
	
	switching modes:
	--------
	[Trains in manual mode]	>> via hotkey
	[Trains in automatic mode] >> via hotkey or mod event trigger:

	mod event trigger: hooked on 'on_train_changed_state with' with change from 'train waiting at station until leave conditions are met' to 'path lost'
	   >> which means: single direction train at a terminal station
	
	Supports:
	-------------------------
	- [Trains in automatic mode] different train composition LL-CCC >> CCC-LL, as long as all locos face the same direction
	- [Trains in manual mode] every train composition (singe and dual direction trains)
	- cargo inventory + fuel inventory including burner fuel bar
	- schedules
	- locomotive colors
	- supports base game locomotive, cargo-wagon, artillery-wagon, fluid-wagon
	- custom hotkey for switching of manual ridden trains
	- trains in automatic mode and manual_mode via hotkey - train_mode 'automatic/manual' is preserved during switching
	- should support all modded loco + wagons which were derived from the base game entity prototypes and are only modded in the entity prototype >> @me: more testing ...!
	- mod settings for locomotive fuel check low-level warning
	- supports equipment grid transfer if vehicle has equipment grid [mod VehicleGrid_0.2.0] >> @me: more testing ...!
	
	>> After every direction change train engineer performs a fuel check on every loco in the handled train. 
	   A message is given on the console @ 'fuel inventory < [mod settings value]' and @ 'fuel inventory = 0',
	   the energy remaining in the burner (fuel bar) is not taken into account
		
		
	Limitations:
	-------------------------
	>> [Trains in automatic mode] only trains with locos facing in one direction will fire Train engineer event
	>> switching trains only at stations with straight rails > no curves or diagonal rails within the length of the train allowed!
	>> no cargo-wagon filters currently supported
	>> equipment grids currently supported, but battery charge is 0 after switch and equipment order is changed
	>> only one driver/passenger stays automatically on the train - remaining passenger have to manual reenter or walk home ;-)
	>> not multiplayer tested or tested on servers!!! - disconnects probable
	>> only standard lenght wagons supported (lenght == 6), short wagons like yuoki's 'zement wagons' will fail on Train creation (displacement + not connected)
	>> only fluid-wagons with 1 fluidboxes are supported
	>> locomotives & carriages with special behavior e.g. electric power locomotives (functions via control.lua) will often lead to crashs or non function after switching
	

	>> To Do's:
	-------------------------
	-- code optimisation + cleanup
	-- Train::id (uint) overflow?! what happens
	-- copy_settings(entity) → dictionary string → uint
	-- cargo-wagon filters?
	-- maybe change the fuel check to calculate the remaining MJ in a loco instead of the remaining quantity of the fuel items
	-- raise_event for carriage.destroy() > inform other mods about removed entity if possible
	-- test compatibility of equipment_grid transfer
	-- seperate func for get_equipment_grid_contents	
	-- MORE TESTING
		
	>> questions @ forum:
	----------------------
	- what happens at a Train::id overflow? is there a way to preserve the old train id?
	- is there a clean way to copy settings like inventory filters from a source carriage to a new carriage while the source entity allready got removed via *.destroy() e.g. via a 'invisible dummy train entity'? 
      As I understand copy_settings() only works with two existing entities.
	- in debug mode "F4" >> show-time-usage >> x/y/z >>> what stand z for? Even on a small map z goes up 'linear' to the lenght of just one test train (lenght=15) with z beyond 20++
	   x/y/z >> is it avg./min./max time? Z this high doesn't seems right! Can anyone point me to informations/guides on how to optimize my code / reduce the time usage?
    - volunteer guinea pig to test the mod are very welcome
	  
	  

	>> test environment:
	-----------------------------
	- factorio 0.16.36 win64 steam version on core i7
	
	
	>> compatibilities & tested mods:
	-----------------------------
	- FARL + FARL module
	- yuokis locos + wagons (as long as carriage length = 6 tiles)
	- LTN
	- Angel's Addons - Petrochem Train (tested with water + steam)
	- Nuclear Locomotives 1.1.4 by GotLag
	- 'shuttle train', because its an equipment grid item
		  
	>> known problems/incompatibilities/bugs:
	------------------------------------------ 
	- 'Diesel Locomotive' by evildogbot100 (due to 'non chemical' liquid fuel) >> crash
	- 'TheFatController' >> see known problems/incompatibilities/bugs
	- 'vehicle wagon' when loaded with a vehicle (loading a vehicle onto it and switching the train works, the following unloading leads to crash)
	- 'Fusion Train' by magu5026 >> crash
	- 'Electric Train'  by magu5026 >> no crash, but switched new E-Locomotive has no fuel/power - only manual building seems to connenct the train entitiy to the power provider
	- 'Senpais Trains' by LuziferSenpai >> crash when switching the battle electronic train mk1-3
		  	
	- TheFatController >> every from TE new created new train adds an entry in a table, which doesn't gets removed when the entity is destroyed by TE.
	   See the increasing table numbers in Fat Gui. Page number increase but only the real currently running trains are listed on the last page!
	   F4 debug: show-time-usage x/y/z >> z of TheFatController constantly goes up with every new train, making the game lagging. Probably the hole table is in 
	   some way calculated even the old TE trains got destroyed!
	   >> t = game.surfaces[1].get_trains() >> #t is constant! No dead entries from TE trains ?!?!
	 
	>> performance:
	-------------------
	  - x/y/z dependent on TE event spread and number of switched trains. Many trains switched at the same time and z goes up 
		>> see performance excel for needed calculation time with different trains
	  
	>> history:
	-----------------------------
	 Build: 1804xx
		- initial build and post in forum
	 
	 Build: 180423 
		- small changes in func build_train
		- testing bug in TheFatController
		
	Build: 180424
		- add + test automatic dismount of train passengers
		- rewrite of func train_build >> pos xy now precalc before call of train_build
	
	Build: 180426:
		- automatic dismount & mount of train passengers (all passengers automatically get dismounted before switching,
	      first found driver/passenger in the train becomes the new driver - remaining passenger have to walk home ;-)
		- a bit code optimization
			   		
	Build: 180427:
		- added custom hotkey for switching of ridden trains
		- added mod settings for fuel-low-level item count + message level
		
	Build: 180428:
		- testing for incompatibilities on different mods
		- added support for cargo-wagon blocked slots
	
	Build: 180501:		
		- added equipment grid support
		
	>>IMPORTANT:
	-----------------------------
        - Backup your game data, saves and mod folder before use!
        - use on your own risk!
		- Beware of the noob programmer >> me ;-)
	  
	>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
	
]]--

-- simple rounding function
function round(float)
	return math.floor(float + 0.5)
end

-- return the length of a table >> todo: check if equal to #table within this mod
function get_table_length(T)
	local count = 0
	for _, _ in pairs(T) do count = count + 1 end
	return count
end

-- returns the direction of the first loco in the train
function get_train_direction(train)

	local direction = nil
	
	for _, carriage in pairs(train) do
		if carriage.type == "locomotive" then		
			if direction == nil then
				direction = round(carriage.orientation * 8)	
				break
				--print_debug(">> Set train_direction: " .. direction)				
			end
		end
	end

	return direction
end

-- returns the fuel of a loco entity (inventory + burner)
function get_loco_fuel(loco)

	if loco == nil then 
		print_debut("get_loco_fuel(loco) >> loco == nil!")
		return nil 
	end
	
	local loco_fuel = {}
	local quantity = 0
	
	loco_fuel[1] = loco.get_fuel_inventory().get_contents() 
	
	if loco_fuel[1] ~= nil then
	
		for i, q in pairs(loco_fuel[1]) do
			quantity = quantity + q
		end		
		loco_fuel[4] = quantity
	end
	
	if loco.burner.currently_burning ~= nil then
		loco_fuel[2] = loco.burner.currently_burning.name
		loco_fuel[3]  = loco.burner.remaining_burning_fuel
	end

	return loco_fuel
end


-- todo >> script.on_init(apply_mod_settings())
script.on_init(function ()
	apply_mod_settings()	
end)

script.on_configuration_changed(function (data)
	apply_mod_settings()		
end)

script.on_event(defines.events.on_runtime_mod_setting_changed,function ()
	apply_mod_settings()
end)

function apply_mod_settings()
	
	if settings.global["setting-fuel-check-low-level"].value then
       global.fuel_low_level = settings.global["setting-fuel-check-low-level"].value
    end

end


--[[
==============================================================================
==============================================================================
DEBUG:
==============================================================================
]]

-- auto trigger for bulk logging
script.on_nth_tick(75, function(event)
	
	if cnt == nil then 
		cnt = 1 
	end	
	if go == nil then go = false end
	
	if go == true and cnt < 101 then 
		game.players[1].print("Loop #: " .. cnt)
		auto_switch_debug()
		cnt = cnt + 1			
	end
	
	if cnt == 100 then 
		cnt = 1
		go = false 
	end

end)

-- event trigger for debug hotkey >> bulk cargo filling + logging
script.on_event("hotkey-switch-train-debug", function(event)

	game.players[1].print("hotkey-switch-train-debug fired @" .. (event.tick))	

	if go == false or go == nil then 
		go = true 
	
		local trains = {}
		trains = game.surfaces[1].get_trains() 

		for i = 1, #trains,1 do
		
			for k, carriage in pairs(trains[i].carriages) do				
				if carriage.type == "cargo-wagon" then
					carriage.insert({name="copper-plate", count=500})	
					carriage.insert({name="iron-plate", count=500})	
					carriage.insert({name="fast-transport-belt", count=500})	
					carriage.insert({name="express-transport-belt", count=500})	
					carriage.insert({name="processing-unit", count=500})	
					carriage.insert({name="plastic-bar", count=500})	
					carriage.insert({name="concrete", count=500})	
					carriage.insert({name="hazard-concrete", count=500})
				end			
			end	
		end
		
		game.players[1].print("start auto log for #" ..#trains)			
		
	end
	
end)

function auto_switch_debug()

	local train_list = {}
	train_list = game.surfaces[1].get_trains() 
	
	log("start_logging: " .. cnt)
		for i = 1, #train_list,1 do
			switch_train(train_list[i])
		end
	log("end_logging" .. cnt) -- get ms 

end


--[[
==============================================================================
DEBUG END
==============================================================================
==============================================================================
]]




-- event trigger for hotkey while riding a train
script.on_event("hotkey-switch-train", function(event)
	
	local vehicle = game.players[event.player_index].vehicle
    if vehicle and vehicle.train then
		switch_train(vehicle.train)
    end
	
end)

-- event trigger for trains in automatic mode (only fired from single headed trains at terminal stations)
script.on_event(defines.events.on_train_changed_state, function(event)
      
	if event.old_state == defines.train_state.wait_station and event.train.state == defines.train_state.path_lost then
		switch_train(event.train)
	end
	
end)


function switch_train(event_train)

		--log("start function switch_train")
		local train_carriages = {}
		local x_inventory = {}		
		local new_schedule = event_train.schedule
		
		local can_build = true
		local created_train
		
		logging = 2 

		local t_carriage_count = get_table_length(event_train.carriages)		
		local t_direction = get_train_direction(event_train.carriages) 
		local t_pos = {x = {}, y = {}}
		local tl = (t_carriage_count * 6) + (t_carriage_count - 1) 
		local carriage_offset = 0
		local mul_x, mul_y = 0
		local t_driver 
		local train_mode = event_train.manual_mode
		local blocked_slots = 0
		local equipment_grid = {}
		
		--game.players[1].clear_console()	
		--print_debug(" ======================================== ")
		--print_debug(" Event fired @" .. event.tick .. " - #" .. event.train.id)		
		--print_debug(" Train Direction: " .. train_direction)		
		--print_debug(" Train lenght: " .. t_carriage_count)
		--print_debug(" Event for Train #" .. event.train.id)		
		--print_debug(" ---------------------------------------- ")
		
		
		-- get first player as future driver & dismount all passengers
		if #event_train.passengers > 0 then
			--print_debug(" Engineer Bob: Get off my Train - I'm working!!!")
	
			for i = 1, t_carriage_count,1 do
			
				if t_driver == nil then
					t_driver = event_train.carriages[i].get_driver()
					--break
				end
				event_train.carriages[i].set_driver(nil)
				
			end	
		end
				

		-- get carriage type, inventory, fuel, color, equipment
		for i, carriage in pairs(event_train.carriages) do	
					
			if carriage.type == "locomotive" then
			
			    if carriage.grid then
					equipment_grid = carriage.grid.equipment --get_contents()
					--print_table(equipment_grid, "equipment_grid")
				else
					--game.players[1].print("no grid found!")
				end
			
				x_inventory = get_loco_fuel(carriage)								
				train_carriages[i] = {carriage_name = carriage.name, carriage_type = carriage.type, inventory = x_inventory, carriage_color = carriage.color, carriage_grid_contents = equipment_grid}

			elseif carriage.type == "cargo-wagon" or carriage.type == "artillery-wagon" then
			
				if carriage.grid then
					equipment_grid = carriage.grid.equipment --get_contents()
					--print_table(equipment_grid, "equipment_grid")
				else
					--game.players[1].print("no grid found!")
				end
			
				x_inventory = carriage.get_inventory(defines.inventory.cargo_wagon)
				if x_inventory.hasbar() then
					blocked_slots = x_inventory.getbar()
				end				
				
				x_inventory = x_inventory.get_contents() 		
				train_carriages[i] = {carriage_name = carriage.name, carriage_type = carriage.type, inventory = x_inventory, carriage_slots = blocked_slots, carriage_grid_contents = equipment_grid}	
		
			elseif carriage.type == "fluid-wagon" then
				
				x_inventory =  carriage.fluidbox[1]				
				train_carriages[i] = {carriage_name = carriage.name, carriage_type = carriage.type, inventory = x_inventory}
										
			end				
			
			-- calculate carriage positions
			if i == 1 then	
				
				if t_direction == defines.direction.north then --if _train_direction == defines.direction.south then
					t_direction = defines.direction.south 
					
					t_pos.x[i] = round(carriage.position.x)
					t_pos.y[i] = round(carriage.position.y) + tl - 6	
								
					carriage_offset = -7
					mul_x = 0
					mul_y = 1
					
				elseif t_direction == defines.direction.south then --elseif _train_direction == defines.direction.north then				
					t_direction = defines.direction.north 
					
					t_pos.x[i] = round(carriage.position.x)
					t_pos.y[i] = round(carriage.position.y) - tl + 6	
					
					carriage_offset = 7
					mul_x = 0
					mul_y = 1
					
				elseif t_direction == defines.direction.east then --elseif _train_direction == defines.direction.west then
					t_direction = defines.direction.west 
					
					t_pos.x[i] = round(carriage.position.x) - tl + 6
					t_pos.y[i] = round(carriage.position.y)
					
					carriage_offset = 7
					mul_x = 1
					mul_y = 0
					
				elseif t_direction == defines.direction.west then --elseif _train_direction == defines.direction.east then
					t_direction = defines.direction.east 
					
					t_pos.x[i] = round(carriage.position.x) + tl - 6
					t_pos.y[i] = round(carriage.position.y)
					
					carriage_offset = -7
					mul_x = 1
					mul_y = 0
					
				else
					print_debug(" Engineer Bob: I'm not working on diagonal rails!")	
					can_build = false
					break
				end
				
			else			
				
				if t_direction == 0 or t_direction == 2 or t_direction == 4 or t_direction == 6 then
					
					t_pos.x[i] = t_pos.x[i-1] + mul_x * carriage_offset
					t_pos.y[i] = t_pos.y[i-1] + mul_y * carriage_offset
				
				else
					print_debug(" Engineer Bob: I'm not working on diagonal rails!")
					can_build = false
					break					
				end				
	
			end
			
		end

		-- remove old train & build + setup new train
		if can_build == true then

			for i, carriage in pairs(event_train.carriages) do	
				carriage.destroy()
			end

			-- todo!!!
			-- for i = t_carriage_count, 1, -1 do	
			   -- -- script.raise_event(defines.events.on_entity_died, {entity=event_train.carriages[i]})
				-- event_train.carriages[i].destroy()
			-- end
			
			
			-- build new train
			created_train = build_train(train_carriages, t_pos, t_direction, t_carriage_count)
			
			-- setup new train (schedule, manuel_mode, driver, ...)
			if created_train ~= nil then
		
				if new_schedule ~= nil then	
					created_train.train.schedule = new_schedule
				end			
								
				if t_driver ~= nil then
					created_train.train.carriages[1].set_driver(t_driver)
				end
				
				created_train.train.manual_mode = train_mode	
			end	
		end
		--log("end function switch_train")
end



-- function to build new train in reversed order with the parameters of the source train (fuel, inventory, ...) 
-- >> inventory is interpreted depending on carriage type (fuel_inventory for loco, cargo_inventory for cargo-wagon, ammunition for arty, fluid)
function build_train(_train_carriages, _pos, _train_direction, _train_lenght)
	
	--log("start::sub-function: build_train")
	local created_entity = nil
	local build_error = false
	local equipment = {}
	--print_debug("Build data >> #cars: " .. _train_lenght .. " Train lenght: " .. tl .. " Dir: " .. _train_direction)
	--print_debug("Build data >> x:" .. _train_x .. " y:" .. _train_y)
	
	for i, new_carriage in pairs(_train_carriages) do	
	
		--created_entity = nil	
		created_entity = game.surfaces[1].create_entity{name=new_carriage.carriage_name, position = {_pos.x[i],_pos.y[i]}, direction = _train_direction, force="player"}

		if created_entity == nil then
			--print_debug(">>> build @: x:" .. _pos[1] .. ";" .. _pos[2] .. " Dir: ".. _train_direction .. " >> Error during creation of this carriage")	
			build_error = true		
		else
			--print_debug(">>> build @: x:" .. _pos[1] .. ";" .. _pos[2] .. " Dir: ".. _train_direction .. " >> Ok")	
				
			if new_carriage.carriage_type == "locomotive" then
			
				if new_carriage.inventory ~= nil then			
					for item, quantity in pairs(new_carriage.inventory[1]) do
						created_entity.insert({name=item, count=quantity})	
					end
						
					if new_carriage.inventory[4] ~= nil then					
						if new_carriage.inventory[4] == 0 then 
							print_debug("Train-Engineer: Locomotive " .. i .. " in Train #" .. created_entity.train.id .. " is out of fuel!")						
						elseif new_carriage.inventory[4] < global.fuel_low_level then 
							print_debug("Train-Engineer: Locomotive " .. i .. " in Train #" .. created_entity.train.id .. " is low on fuel! >> " .. new_carriage.inventory[4])		
						end
					end	
						
					if new_carriage.inventory[2] ~= nil and new_carriage.inventory[3] ~= nil then
						created_entity.burner.currently_burning = new_carriage.inventory[2]
						created_entity.burner.remaining_burning_fuel = new_carriage.inventory[3]						
					end	
				end
				
				--print_table(new_carriage.carriage_color, "carriage_color3")
				
				if new_carriage.carriage_color ~= nil then
					created_entity.color = new_carriage.carriage_color 
				else
					--print_debug("loco is default color!")
				end
			
		
			elseif new_carriage.carriage_type == "cargo-wagon" or new_carriage.carriage_type == "artillery-wagon" then
			
				if new_carriage.inventory ~= nil then			
					for item, quantity in pairs(new_carriage.inventory) do
						created_entity.insert({name=item, count=quantity})	
					end				
				end
				
				x_inventory = created_entity.get_inventory(defines.inventory.cargo_wagon)
				if x_inventory.hasbar() then
					x_inventory.setbar(new_carriage.carriage_slots)
				end	
	
			elseif new_carriage.carriage_type == "fluid-wagon" then
				if new_carriage.inventory ~= nil then
							
					if new_carriage.inventory then
						created_entity.fluidbox[1] = new_carriage.inventory
					end	
					-- ...
				end			
			
			end
			
			if created_entity.grid then
				if new_carriage.carriage_grid_contents then 
				
					for i = 1, #new_carriage.carriage_grid_contents, 1 do
						equipment = created_entity.grid.put{name=new_carriage.carriage_grid_contents[i].name}
						if equipment then
							--game.players[1].print("inserted: " .. equipment.name)
						end
					end
					
				else
					--game.players[1].print("no carriage_grid_contents fount to insert!")
				end
			else
				--game.players[1].print("no grid found on new train!")
			end
			
			
			
		end			
	end
	
	--log("end::sub-function: build_train")
	
	if build_error == false then
		return created_entity
	else
		return nil	
	end	
	
end



----------------------------------------------------------------------------------------------
